#include "minicon.h"

miniCon::miniCon(const char *devicename) {
	if (devicename!=NULL) {
		strcpy(device, devicename);
	}
	else {
		strcpy(device, "/dev/modem");
	}
	iIsOpen=false;
	bitrate=0;
	time=0;
}

/********************
 * OPEN SERIAL PORT *
 ********************/
void miniCon::openDevice() { 
	fd = open(device, O_RDWR | O_NOCTTY | O_NDELAY);
	iIsOpen=true;
	if (fd == -1) {
		#ifdef _DEBUG
     		fprintf(stderr, "unable to open \"%s\" - %s\n",device,strerror(errno));
		#endif
		iIsOpen=false;
	}
	initDevice();
}

/*********************
 * CLOSE SERIAL PORT *
 *********************/
void miniCon::closeDevice(void) {
	close(fd);
	iIsOpen=false;
}

tStatus miniCon::getLastStatus(void) {
	return(iStatus);
}

void miniCon::setTimeout(int sec) {
	time=sec;
}

/*******************************************
 * SET PROPER BAUD RATE AND TERMINAL FLAGS *
 *******************************************/
void miniCon::initDevice(void) {
	
	struct termios options;
	
	if ( !(iIsOpen) ) {
		return;
	}
	
	fcntl(fd, F_SETFL, FNDELAY);
	tcgetattr(fd, &options);
	options.c_cflag |=  (CS8 | CRTSCTS | CREAD | CLOCAL);
	options.c_cflag &= ~(CSTOPB | PARENB);
	
	bitrate=detectBaudrate(options);
	
        tcflush(fd, TCIOFLUSH);
}

/**************************
 * SENDS STRING TO DEVICE *
 **************************/
 
void miniCon::sendString(const char *string) {
	if ( (!(iIsOpen)) || (bitrate==0) ) {
		return;
	}
	write(fd, string, strlen(string));
	write(fd, "\r", 1);
}

/***********************************
 * WAITS FOR STRING TO BE RETURNED *
 ***********************************/

int miniCon::waitString(const char *str1, const char *str2) {
	
	char tmp[128]="\0";
	char received;
	
	while((received=getChar())!=0) {
		if ( (received=='\r') || (received=='\n') ) { 
				strcpy(tmp, "\0");
		}
		else {
			strncat(tmp, &received, 1);
		}
		if ( strlen(tmp)>0 ) {
			if (str1!=NULL) {
				if (strcmp(tmp, str1)==0) {
					#ifdef _DEBUG
					printf("\nfound string 1: %s\n", tmp);
					#endif
					return(1);
				}
			}
			if (str2!=NULL) {
				if (strcmp(tmp, str2)==0) {
					#ifdef _DEBUG
					printf("\nfound string 2: %s\n", tmp);
					#endif
					return(2);
				}
			}
		}	
	} 
	return(0);
}

/***********************************************/
/* CHECKS IF CONNECTED DEVICE IS A SpeedDragon */
/***********************************************/
bool miniCon::checkDragon(void) {
	sendString("ATI1");
	switch (waitString("Speed Dragon", "ERROR")) {
		case 0:
		case 2:
			return(false);
		case 1: 
			return(true);
	}
	return(false);
}

/***************************************
 * PROBES BAUDRATE OF CONNECTED DEVICE *
 ***************************************/ 
long miniCon::detectBaudrate(struct termios terminal) {

	speed_t tbaud=B115200;
	char test='a', bucket;
	long ibaud=115200;
	struct timeval timeout;
	fd_set read_fds;

	FD_ZERO(&read_fds);
	do {
		timeout.tv_sec=0;
		timeout.tv_usec=400000;
		FD_SET(fd, &read_fds);
		select(fd+1,&read_fds,NULL,NULL,&timeout);
		read(fd,&bucket,1);
		#ifdef _DEBUG
		printf(".");
		#endif
	} while (FD_ISSET(fd, &read_fds));
	
	while(1) {		
		timeout.tv_sec=0;
		timeout.tv_usec=180000;
		cfsetispeed(&terminal,tbaud);
		cfsetospeed(&terminal,tbaud);
		tcsetattr(fd, TCSAFLUSH, &terminal);
		#ifdef _DEBUG
		printf("bitrate: %ld...",ibaud);
		#endif
		
		write(fd, &test, sizeof(char));
		FD_SET(fd, &read_fds);
		select(fd+1,&read_fds, NULL, NULL, &timeout);
		if (FD_ISSET(fd, &read_fds)) {
			read(fd, &test , sizeof(char));
			if (test!='a') continue;
			#ifdef _DEBUG
			printf("ok\n");
			#endif
			iStatus=sReady;
			return(ibaud);
		}
		else {
			#ifdef _DEBUG
			printf("failed\n");
			#endif
			switch(ibaud) {
				case 230400:
					ibaud=detectNAK(terminal);
					if (ibaud==0) {
						iStatus=sUnknownError;
						return(0);
					}
					else {
						iStatus=sWaitingForTransfer;
						return(ibaud);
					}
					break;
				case 115200: tbaud=B38400;
					     ibaud=38400;
					break;
				case 57600: tbaud=B9600;
					    ibaud=9600;
					break;
				case 38400: tbaud=B57600;
					    ibaud=57600;
					break;
				case 19200: tbaud=B230400;
					    ibaud=230400;
					break;
				case 9600: tbaud=B19200;
					   ibaud=19200;
					break;
				default: tbaud=B115200;
					 ibaud=115200;
			}
		}
	}
	return(ibaud);
}

/*****************************************************************
 * PROBES BAUDRATE OF CONNECTED DEVICE BY DETECTING NAK (XMODEM) *
 *****************************************************************/ 
long miniCon::detectNAK(struct termios terminal) {

	speed_t tbaud=B115200;
	char test;
	long ibaud=115200;
	struct timeval timeout;
	fd_set read_fds;
	
	FD_ZERO(&read_fds);
	while(1) {		
		timeout.tv_sec=1;
		timeout.tv_usec=300000;
		cfsetispeed(&terminal,tbaud);
		cfsetospeed(&terminal,tbaud);
		tcsetattr(fd, TCSAFLUSH, &terminal);
		#ifdef _DEBUG
		printf("detecting NAK on %ld bps...",ibaud);
		#endif
		
		FD_SET(fd, &read_fds);
		select(fd+1,&read_fds, NULL, NULL, &timeout);
		if (FD_ISSET(fd, &read_fds)) {
			read(fd, &test , sizeof(char));
			if (test==21) {
				#ifdef _DEBUG
				printf("ok\n");
				#endif
				iStatus=sWaitingForTransfer;
				return(ibaud);
			}
			else {
				#ifdef _DEBUG
				printf("failed\n");
				#endif
				switch(ibaud) {
					case 230400:
						return(0); 
						iStatus=sUnknownError;
						break;
					case 115200: tbaud=B38400;
						     ibaud=38400;
						break;
					case 57600: tbaud=B9600;
					 	    ibaud=9600;
						break;
					case 38400: tbaud=B57600;
					    	    ibaud=57600;
						break;
					case 19200: tbaud=B230400;
					    	    ibaud=230400;
						break;
					case 9600: tbaud=B19200;
						   ibaud=19200;
						break;
					default: tbaud=B115200;
						 ibaud=115200;
				}	
			}
		}
		else { /* no char detected */
			#ifdef _DEBUG
			printf("failed - no response\n");
			#endif
			iStatus=sUnknownError;
			return(0);
		}	
	}
}

/***********************************
 * RETURNS NEXT CHAR FROM TERMINAL *
 ***********************************/
char miniCon::getChar(void) {
	
	struct timeval timeout;
	fd_set read_fds;
	char received;
	
	FD_ZERO(&read_fds);
	timeout.tv_sec=time;
	timeout.tv_usec=300000;
	FD_SET(fd, &read_fds);
	select(fd+1,&read_fds,NULL,NULL,&timeout);
	if (FD_ISSET(fd, &read_fds)) {
		read(fd, &received, sizeof(received));
		#ifdef _DEBUG
		printf("%c", received);
		#endif
		return(received);
	}
	else {
		#ifdef _DEBUG
		printf("timeout!\n");
		#endif
		return(0);
	}
}
