//
// connect.cpp
//
// (c) Jan Willamowius, 1998
//
// free for private use - commercial use or commercial distribution
// requires permission from the author
//

#include "connect.hpp"
#include <iostream>
#define __need_timeval
#include <time.h>		// time(), difftime()
#include <ctype.h>		// isprint()

#ifdef linux
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>		// open() und Modi 
#include <unistd.h>		// close(), read(), write()
#include <sys/ioctl.h>	// ioctl()
#include <sys/types.h>
#include <sys/stat.h>	// chmod()
#include <signal.h>		// kill()
#include <termios.h>
#include <pwd.h>
#include <errno.h>
#endif
 
#ifdef _MSC_VER
using namespace std;
#endif
 

ocConnection::ocConnection(const string & sNewDevice)
{
	sDevice = sNewDevice;
	bEcho = false;
	sInitStr = "ATZ\r";
	iBps = 9600;
	bOpenStatus = false;
	iTimeout = 10;		// 10 sec
	iWhichString = 0; 	// none
#ifdef linux
	iDeviceHandle = -1;
#endif
#ifdef _MSC_VER
	idComDev = 0;
#endif
};

ocConnection::~ocConnection()
{
	if (bIsOpen())
	{
		vClose();
	};
};

void ocConnection::vOpen(void)
{
#ifdef _MSC_VER
	DCB dcb;
#endif

	if (bDeviceLocked())
		return;
	
	vLockDevice();

#ifdef linux
	iDeviceHandle = open(sDevice.c_str(), O_RDWR | O_NDELAY);
	if (iDeviceHandle >= 0)
	{  
		bOpenStatus = true;
 
		fcntl(iDeviceHandle, F_SETFL, O_RDWR);
		ioctl(iDeviceHandle, TCGETA, &xtermio);
		xtermio.c_iflag = ICRNL | IXANY;
		xtermio.c_oflag = OPOST | ONLCR;
		xtermio.c_cflag = (unsigned short)(B19200 | CS8 | CREAD | HUPCL | CLOCAL | CRTSCTS);
		xtermio.c_lflag = 0; /* echo off, no signals */
		xtermio.c_line = N_TTY;
		xtermio.c_cc[VMIN] = 1;
		xtermio.c_cc[VTIME] = 0;
		ioctl(iDeviceHandle, TCSETAF, &xtermio);
#endif
 
#ifdef _MSC_VER
   if ((idComDev =
       CreateFile(sDevice.c_str(), GENERIC_READ | GENERIC_WRITE,
                  0,                    // exclusive access
                  NULL,                 // no security attrs
                  OPEN_EXISTING,
                  FILE_ATTRIBUTE_NORMAL,
                  NULL)) != (HANDLE) -1)
	{  
		bOpenStatus = true;

		// setup device buffers
		SetupComm(idComDev, 1024, 1024);
 
		// purge any information in the buffer
		PurgeComm(idComDev, PURGE_TXABORT | PURGE_RXABORT |
                            PURGE_TXCLEAR | PURGE_RXCLEAR);
 
		dcb.DCBlength = sizeof( DCB);
		GetCommState(idComDev, &dcb);
		dcb.ByteSize = 8;
		dcb.Parity = NOPARITY;
		dcb.StopBits = ONESTOPBIT;
		// setup hardware flow control
		dcb.fDtrControl = DTR_CONTROL_HANDSHAKE;
		dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
		// other various settings
		dcb.fBinary = TRUE;
		dcb.fParity = TRUE;
		SetCommState(idComDev, &dcb);
#endif

		vSetBps(iBps);

		vSendString(sInitStr);
		vWaitString("OK");
		if (iStringFound() != 1)
			vClose();
	};
};

void ocConnection::vClose(void)
{
	if (!bIsOpen())
		return;

#ifdef linux
	close(iDeviceHandle);
	iDeviceHandle = -1;
#endif
 
#ifdef _MSC_VER
   // disable event notification and wait for thread
   // to halt
	SetCommMask(idComDev, 0);
 
	// drop DTR
	EscapeCommFunction(idComDev, CLRDTR);
 
	// purge any outstanding reads/writes and close device handle
	PurgeComm(idComDev, PURGE_TXABORT | PURGE_RXABORT |
                        PURGE_TXCLEAR | PURGE_RXCLEAR );
	CloseHandle(idComDev);
 
	idComDev = (HANDLE) -1;
#endif
	bOpenStatus = false;
	vUnlockDevice();
};

void ocConnection::vConnect(const string & sDialStr)
{
	if (!bIsOpen())
		return;

	vSendString(sDialStr);
	vWaitString("CONNECT");
};

void ocConnection::vHangup(void)
{
	if (!bIsOpen())
		return;

	vSendString("+++");
	vWaitString("OK");
	vSendString("ATH\r");
	vWaitString("OK");
};

bool ocConnection::bIsOpen(void) const
{
	return bOpenStatus;
};

void ocConnection::vSetInitStr(const string & sNewInitStr)
{
	sInitStr = sNewInitStr;
};

string ocConnection::sGetInitStr(void) const
{
	return sInitStr;
};

// require: iNewBps must be a valid speed (see list below)
void ocConnection::vSetBps(const int iNewBps)
{
#ifdef linux
	int iSpeedFlag;
 
	iBps = iNewBps;

	// if the device is already open reset bps setting
	// otherwise vOpen() will set it
	if (bIsOpen())
	{
		ioctl(iDeviceHandle, TCGETA, &xtermio);
		switch (iBps)
		{
			case 460800: iSpeedFlag = B460800;
						 break;
			case 230400: iSpeedFlag = B230400;
						 break;
			case 115200: iSpeedFlag = B115200;
						 break;
			case 57600 : iSpeedFlag = B57600;
						 break;
			case 38400 : iSpeedFlag = B38400;
						 break;
			case 19200 : iSpeedFlag = B19200;
						 break;
			case  9600 : iSpeedFlag = B9600;
						 break;
			case  4800 : iSpeedFlag = B4800;
						 break;
			case  2400 : iSpeedFlag = B2400;
						 break;
			case  1200 : iSpeedFlag = B1200;
						 break;
			case   300 : iSpeedFlag = B300;
						 break;
			default    : iSpeedFlag = B9600;
						 break;
		}
		xtermio.c_cflag = iSpeedFlag | CS8 | CREAD | HUPCL | CLOCAL | CRTSCTS;
		ioctl(iDeviceHandle, TCSETAF, &xtermio);
	};
#endif

#ifdef _MSC_VER
	DCB dcb;

	iBps = iNewBps;

	// if the device is already open reset bps setting
	// otherwise vOpen() will set it
	if (bIsOpen())
	{
		dcb.DCBlength = sizeof( DCB);
		GetCommState(idComDev, &dcb);

		switch (iBps)
		{
/*
			case 460800: dcb.BaudRate = CBR_460800;
						 break;
			case 230400: dcb.BaudRate = CBR_230400;
						 break;
*/
			case 115200: dcb.BaudRate = CBR_115200;
						 break;
			case 57600 : dcb.BaudRate = CBR_57600;
						 break;
			case 38400 : dcb.BaudRate = CBR_38400;
						 break;
			case 19200 : dcb.BaudRate = CBR_19200;
						 break;
			case  9600 : dcb.BaudRate = CBR_9600;
						 break;
			case  4800 : dcb.BaudRate = CBR_4800;
						 break;
			case  2400 : dcb.BaudRate = CBR_2400;
						 break;
			case  1200 : dcb.BaudRate = CBR_1200;
						 break;
			case   300 : dcb.BaudRate = CBR_300;
						 break;
			default    : dcb.BaudRate = CBR_9600;
						 break;
		}
		SetCommState(idComDev, &dcb);
	};
#endif
};

void ocConnection::vAutoSetBps(void)
{
#ifdef linux
	/* thanks to Joachim Dorner <jdorner@dillingen.baynet.de> */
	/* for this nice feature */

	speed_t tbaud=B9600;
	char test='a', bucket;
	long ibaud=9600;
	struct timeval timeout;
	fd_set read_fds, write_fds;
	struct termios terminal;

	if (bDeviceLocked())
		return;

	iDeviceHandle = open(sDevice.c_str(), O_RDWR | O_NDELAY);
	tcgetattr(iDeviceHandle, &terminal);
	terminal.c_lflag &= ~(ICANON | ECHO | ISIG);
	tcsetattr(iDeviceHandle, TCSAFLUSH, &terminal);
	
	FD_ZERO(&read_fds);
	FD_ZERO(&write_fds);
	#ifdef DEBUG
	cout << "flushing buffer:" << endl;
	#endif
	do {
		timeout.tv_sec=0;
		timeout.tv_usec=400000;
		FD_SET(iDeviceHandle, &read_fds);
		select(iDeviceHandle+1,&read_fds,NULL,NULL,&timeout);
		read(iDeviceHandle,&bucket,1);
		#ifdef DEBUG
		cout << ".";
		#endif
	} while (FD_ISSET(iDeviceHandle, &read_fds));
	#ifdef DEBUG
	cout << "done" << endl << "probing bitrate:" << endl;
	#endif
	while(1) {
		timeout.tv_sec=0;
		timeout.tv_usec=180000;
		cfsetispeed(&terminal,tbaud);
		cfsetospeed(&terminal,tbaud);
		tcsetattr(iDeviceHandle, TCSAFLUSH, &terminal);
		#ifdef DEBUG
		cout << "\t" << ibaud << "...";
		#endif
		write(iDeviceHandle,&test,sizeof(test));
		FD_SET(iDeviceHandle, &read_fds);
		select(iDeviceHandle+1,&read_fds,NULL,NULL,&timeout);
		if (FD_ISSET(iDeviceHandle, &read_fds)) {
			read(iDeviceHandle,&test,sizeof(test));
			if (test!='a') continue;
			#ifdef DEBUG
			cout << "ok" << endl;
			#endif
			close(iDeviceHandle);
			iBps = ibaud;
			return;
		}
		else {
			#ifdef DEBUG
			cout << "failed" << endl;
			#endif
			switch(ibaud) {
				case 230400:iBps=0;
							// test failed, return
							close(iDeviceHandle);
							return;
							break;
				case 115200:tbaud=B230400;
							ibaud=230400;
							break;
				case 57600: tbaud=B115200;
					    ibaud=115200;
					break;
				case 38400: tbaud=B57600;
					    ibaud=57600;
					break;
				case 19200: tbaud=B38400;
					    ibaud=38400;
					break;
				case 9600: tbaud=B19200;
					   ibaud=19200;
					break;
				default: tbaud=B9600;
					 ibaud=9600;
			}
		}
	}
#endif
}

int ocConnection::iGetBps(void) const
{
	return iBps;
}

void ocConnection::vSetEcho(const bool bNewEcho)
{
	bEcho = bNewEcho;
}

bool ocConnection::bGetEcho(void) const
{
	return bEcho;
}

void ocConnection::vSetTimeout(const int iNewTimeout) // je Zeichen ?
{
	iTimeout = iNewTimeout;
};

int ocConnection::iGetTimeout(void) const
{
	return iTimeout;
};

void ocConnection::vSendChar(const char ch) const
{
#ifdef linux
	int iCount;
#endif
#ifdef _MSC_VER
	DWORD iCount;
	BOOL fWriteStat;
#endif
	time_t xStartTime;

//	if (!bIsOpen())
//		return;

	xStartTime = time(NULL);
	do
	{
#ifdef linux
		iCount = write(iDeviceHandle, &ch, 1);
#endif
#ifdef _MSC_VER
		fWriteStat = WriteFile(idComDev, &ch, 1, &iCount, NULL);
#endif
 
	} while ((iCount != 1) && (difftime(time(NULL), xStartTime) < iTimeout));
};

void ocConnection::vSendString(const string & sSt) const
{
	unsigned int iPos;

	for (iPos=0; iPos < sSt.length(); iPos++)
		vSendChar(sSt[iPos]);
};

void ocConnection::vReceiveChar(void)
{
	char ch;
	time_t xStartTime;
#ifdef _MSC_VER
	DWORD iCount = 0;
#endif

#ifdef linux
	int iCount = 0;
	timeval xTimeVal;
	fd_set xReadSet;
#endif

	if (!bIsOpen())
		return;

#ifdef linux
	xTimeVal.tv_sec = iTimeout;
	xTimeVal.tv_usec = 0;
	FD_ZERO(&xReadSet);
	FD_SET(iDeviceHandle, &xReadSet);
#endif
	xStartTime = time(NULL);

	do
	{
		// old: blocking IO
		// iCount = read(iDeviceHandle, &ch, 1);
#ifdef _MSC_VER
		ReadFile(idComDev, &ch, 1, &iCount, NULL);
#endif
 
#ifdef linux
		// new: use select to wait a limited time
		if (select(iDeviceHandle+1, &xReadSet, NULL, NULL, &xTimeVal) > 0)
		{
			// data available
			iCount = read(iDeviceHandle, &ch, 1);
			if (bEcho && (iCount == 1)) {
				if (isprint(ch)) {
					cout << ch << flush;
				} else {
					cout << '.' << flush;
				};
			};
		};
#endif
	} while ((iCount != 1) && (difftime(time(NULL), xStartTime) < iTimeout));
	chLastChar = ch;
};

void ocConnection::vReceiveString(void)
{
	string sBuffer = "";
	time_t xStartTime;
#ifdef _MSC_VER
	const char cEol = '\r';
#else
	const char cEol = '\n';
#endif

	xStartTime = time(NULL);
	do
	{
		vReceiveChar();
		if (chGetLastChar() != cEol)
			sBuffer += chGetLastChar();	// Timeout beruecksichtigen ??
	} while ((chGetLastChar() != cEol) &&
		   (difftime(time(NULL), xStartTime) < iTimeout));
	sLastString = sBuffer;
};

// require: vorher muss ReceiveChar aufgerufe worden sein
char ocConnection::chGetLastChar(void) const
{
	return chLastChar;
}

// require: vorher muss ReceiveString aufgerufen worden sein
string ocConnection::sGetLastString(void) const
{
  return sLastString;
}

void ocConnection::vWaitString(const string & sSt1)
{
	time_t xStartTime;
	string sSoFarReceived = "";

	iWhichString = 0; 	// None
	xStartTime = time(NULL);
	while ((difftime(time(NULL), xStartTime) < iTimeout) && (iWhichString == 0))
	{
		vReceiveChar();
		sSoFarReceived += chGetLastChar();
		if (sSoFarReceived.find(sSt1) != string::npos)
			iWhichString = 1;
	}
};

void ocConnection::vWaitString(const string & sSt1, const string & sSt2)
{
	time_t xStartTime;
	string sSoFarReceived = "";

	iWhichString = 0; 	// None
	xStartTime = time(NULL);
	while ((difftime(time(NULL), xStartTime) < iTimeout) && (iWhichString == 0))
	{
		vReceiveChar();
		sSoFarReceived += chGetLastChar();
		if (sSoFarReceived.find(sSt1) != string::npos)
			iWhichString = 1;
		if (sSoFarReceived.find(sSt2) != string::npos)
			iWhichString = 2;
	}
};

void ocConnection::vWaitString(const string & sSt1, const string & sSt2, const string & sSt3)
{
	time_t xStartTime;
	string sSoFarReceived = "";

	iWhichString = 0; 	// None
	xStartTime = time(NULL);
	while ((difftime(time(NULL), xStartTime) < iTimeout) && (iWhichString == 0))
	{
		vReceiveChar();
		sSoFarReceived += chGetLastChar();
		if (sSoFarReceived.find(sSt1) != string::npos)
			iWhichString = 1;
		if (sSoFarReceived.find(sSt2) != string::npos)
			iWhichString = 2;
		if (sSoFarReceived.find(sSt3) != string::npos)
			iWhichString = 3;
	}
};

void ocConnection::vWaitString(const string & sSt1, const string & sSt2, const string & sSt3, const string & sSt4)
{
	time_t xStartTime;
	string sSoFarReceived = "";

	iWhichString = 0; 	// None
	xStartTime = time(NULL);
	while ((difftime(time(NULL), xStartTime) < iTimeout) && (iWhichString == 0))
	{
		vReceiveChar();
		sSoFarReceived += chGetLastChar();
		if (sSoFarReceived.find(sSt1) != string::npos)
			iWhichString = 1;
		if (sSoFarReceived.find(sSt2) != string::npos)
			iWhichString = 2;
		if (sSoFarReceived.find(sSt3) != string::npos)
			iWhichString = 3;
		if (sSoFarReceived.find(sSt4) != string::npos)
			iWhichString = 4;
	}
};

int ocConnection::iStringFound(void) const  // Which string was found ?
{
	return iWhichString;
};

string ocConnection::sGetLockfileName(void)
{
#ifdef _MSC_VER
	return "";
#endif
#ifdef linux
	string sLockFile = "/var/lock/LCK..";
	int i;

        for (i=sDevice.size()-1; i>=0; i--) {
                if (sDevice[i]=='/') {
                        i++;
                        break;
                }
        }
	sLockFile.append(sDevice.c_str()+i);

	return sLockFile;
#endif
};

bool ocConnection::bDeviceLocked(void)
{
#ifdef _MSC_VER
	return false;
#endif
 
#ifdef linux
	int iFd;
	char cBuf[10];
	int iPid;

	if ((iFd = open(sGetLockfileName().c_str(), O_RDONLY | O_NONBLOCK)) >=0 )
	{
		read(iFd, &cBuf, 10);
		close(iFd);
		// check for stale lock file
		iPid = atoi(cBuf);

		if (iPid == getpid())
			return false;	/* _our_ lockfile */

		if ((kill(iPid, 0) == -1) && (errno == ESRCH))
		{
			// stale lock
			unlink(sGetLockfileName().c_str());
			return false;
		}
		return true;
	}
	else
		return false;
#endif
}

void ocConnection::vLockDevice(void)
{
#ifdef linux
	struct passwd *pwdEntry = NULL;
	FILE* xFd = NULL;

	pwdEntry=getpwuid(geteuid());
	xFd = fopen(sGetLockfileName().c_str(), "w");
	if (xFd)
	{
		fprintf(xFd, "%i connect %s\n", getpid(), pwdEntry->pw_name);
		fclose(xFd);
	}
	chmod(sGetLockfileName().c_str(), 0644);
#endif
}

void ocConnection::vUnlockDevice(void)
{
#ifdef linux
	remove(sGetLockfileName().c_str());
#endif
}

int ocConnection::iGetFileDescriptor(void)
{
#ifdef _MSC_VER
	return 0;
#endif
#ifdef linux
	return iDeviceHandle;
#endif
}

