#include "rcfile.h"

rcFile::rcFile(const char* filename, const char *path) {
	iIsOpen=false;
	iIsValid=false;
	if (path!=NULL) {
		iFileName=path;
		iFileName+="/";
	}
	iFileName+=filename;
	rootEntry=NULL;
}

rcFile::~rcFile(void) {
	rcFileEntry *currentEntry=rootEntry;
	
	if (iIsOpen)
		closeRcFile();
	if (rootEntry==NULL) {
		return;
	}
	else {
		while(currentEntry!=NULL) {
			currentEntry=rootEntry->nextEntry;
			delete rootEntry;
			rootEntry=currentEntry;
		}
	}
}

/* reads rc-file and puts entries in a single chained list.
   if checksum is true a checksum will be expected. if there's none
   isValid() will return false. The checksum will be deleted from the entry 
   list.
   returns NULL if file couldn't be opened */
bool rcFile::readRcFile(bool checksum) {
	
	char *line;
	char tmpLine[MAX_LINE_LENGTH+1];
	rcFileEntry *tmpEntry, *listEntry;
	
	if (!openRcFile(true))  	/* read-only */
		return false;
	
	iIsValid=true;
	
	/* test if rcFile is being reread - so the list will have to be deleted */
	if (rootEntry!=NULL) {
		listEntry=rootEntry;
		while(listEntry!=NULL) {
			listEntry=rootEntry->nextEntry;
			delete rootEntry;
			rootEntry=listEntry;
		}
	}
	
	/* creating the root */
	if ((line=readNextItem())!=NULL) {	
		listEntry=new rcFileEntry;
		strcpy(tmpLine, line);
		tmpEntry=parseLine(tmpLine);
		strcpy(listEntry->key, tmpEntry->key);
		strcpy(listEntry->value, tmpEntry->value);
		listEntry->type=tmpEntry->type;
		listEntry->nextEntry=NULL;
		rootEntry=listEntry;
	}
	else	{
		iIsValid=false; /* file read - but empty */
		closeRcFile();
		return(true); 
	}
		
	/* adding entries to the list */
	while ((line=readNextItem())!=NULL) {
		strcpy(tmpLine, line);
		listEntry->nextEntry=new rcFileEntry; /* alloc the next entry */
		listEntry=listEntry->nextEntry; /* step forward */
		tmpEntry=parseLine(tmpLine); /* parse line got by readNextItem() */
		strcpy(listEntry->key, tmpEntry->key); 
		strcpy(listEntry->value, tmpEntry->value);
		listEntry->type=tmpEntry->type;
		listEntry->nextEntry=NULL;	
	}
	
	closeRcFile();
	
	/* if there's a checksum, check and delete it */
	listEntry=rootEntry;
	tmpEntry=NULL;
	if (listEntry!=NULL) {
		/* move to end of list */
		while (listEntry->nextEntry!=NULL) { 
			tmpEntry=listEntry;
			listEntry=listEntry->nextEntry;		
		}
		/* look if there's a checksum */
		if (strcmp(listEntry->key, "@checksum@")!=0) { /* no checksum available */
			if (checksum==true) {
				iIsValid=false;
				#ifdef _DEBUG
				cout << "no checksum included!" << endl;
				#endif
			}
		}
		else {
			if (tmpEntry!=NULL)
				tmpEntry->nextEntry=NULL; /* remove checksum from list */
			if (strcmp(listEntry->value, calcChecksum())==0) {
				iIsValid=true;
				#ifdef _DEBUG
				cout << "checksum valid!" << endl;
				#endif
			}
			else {
				iIsValid=false;
				#ifdef _DEBUG
				cout << "checksum invalid!" << endl;
				cout << "got: \"" << listEntry->value << "\" " 
					"expected: \"" << calcChecksum() << "\"" << endl; 
				#endif 
			}
			delete listEntry;
		}
	}
	else { /* empty file */
		iIsValid=false;
		cout << "file empty!" << endl;
	}
	
	#ifdef _DEBUG
	cout << endl;
	#endif
	return(true);
}

/* purges rc-file by deleting all empty sections. Additionally you can specify
   if empty lines, comments and unknown lines should be removed, too. 
   note that if the cursor pointed to an entry taht will be deleted, 
   it will be reset to point to the very first entry in the list */
void rcFile::purgeRcFile(bool delNL, bool delComments, bool delUnknown) {
	
	rcFileEntry *currentEntry, *tmpEntry;
	bool bDelete;
	int line=1;
	
	if (rootEntry==NULL) { /* list is empty */
		return;
	}

	currentEntry=rootEntry;
	tmpEntry=rootEntry;
	
	while (currentEntry!=NULL) {
		bDelete=false;
		switch (currentEntry->type) {
			case tlComment:
				if (delComments) {
					bDelete=true;
					#ifdef _DEBUG
					cout << "purging comment: line " << line;
					#endif
				}
				break;
			case tlEmpty:
				if (delNL) {
					bDelete=true;
					#ifdef _DEBUG
					cout << "purging NL: line " << line;
					#endif
				}
				break;
			case tlUnknown:
				if (delUnknown) {
					bDelete=true;
					#ifdef _DEBUG
					cout << "purging unknown string: line " << line;
					#endif
				}
				break;
			case tlSection:
				if (countKeys(currentEntry->key)==0) {
					bDelete=true;
					#ifdef _DEBUG
					cout << "purging section header: line " << line;
					#endif
				}
				else {
					bDelete=false;
				}
				break;
			default:
				bDelete=false;
				break;
		}
		if (bDelete==true) {
			if (currentEntry==rootEntry) {
				rootEntry=rootEntry->nextEntry;
				if (currentEntry==cursor) {
					#ifdef _DEBUG
					cout << "purge: reset cursor!" << endl;
					#endif
					cursor=rootEntry;
				}
				delete currentEntry;
				currentEntry=rootEntry;
				tmpEntry=rootEntry;
			}
			else {
				tmpEntry->nextEntry=currentEntry->nextEntry;
				if (currentEntry==cursor) {
					#ifdef _DEBUG
					cout << "purge: reset cursor!" << endl;
					#endif
					cursor=rootEntry;
				}
				delete currentEntry;
				currentEntry=tmpEntry;
			}
			#ifdef _DEBUG
			cout << "...done" << endl;
			#endif
		}
		else {
			tmpEntry=currentEntry;
			currentEntry=currentEntry->nextEntry;
		}
		line++;
	}
}

/* saves entries out of chained list to the rc-file.
   if checksum is true a checksum will be appended to the file.
   returns NULL if file couldn't be opened */
bool rcFile::writeRcFile(bool checksum=false) {
	
	rcFileEntry *currentEntry=rootEntry;
	char tmpLine[MAX_LINE_LENGTH+1];
	if (!openRcFile(false)) /* write-only */
		return false;
	
	iIsValid=true;
	
	currentEntry=rootEntry;
	while (currentEntry!=NULL) {
		if (currentEntry->type==tlEmpty) { /* if line is empty */
			strcpy(tmpLine, "\0");
		}
		else if(currentEntry->type==tlSection) { /* if it's a section */
			sprintf(tmpLine, "%s" , currentEntry->key);
		}
		else if(currentEntry->type==(tlComment)) { /* if it's a comment */
			sprintf(tmpLine, "%s" , currentEntry->key);
		}
		else if(currentEntry->type==(tlKey)) { /* if it's unknown */
			sprintf(tmpLine, "%s = %s", currentEntry->key,
				currentEntry->value);
		}
		else {
			sprintf(tmpLine, "%s" , currentEntry->key);
		}
		writeNextItem(tmpLine);
		currentEntry=currentEntry->nextEntry;	
	}
	
	if (checksum==true) { /* append checksum to file */
		sprintf(tmpLine, "@checksum@ = %s", calcChecksum());
		writeNextItem(tmpLine);
	}
	
	closeRcFile();
	return(true);
}

/* search a key in a given section. this allows to use equal key names in
   different sections. 
   returns key's value or NULL if not found (or file couldn't be opened) */
char* rcFile::getKey(const char *section, const char* key) {
	
	char *sectionItem;
	rcFileEntry *currentEntry=rootEntry;
	
	/* test if section exists */
	if ((currentEntry=getSection(section))==NULL) {
		return(NULL);
	}
	
	/* search matching key */
	sectionItem=getItem(key, currentEntry->nextEntry);
	
	/* check if found key is inside given section */
	if (iEnteredNewSection) {
		#ifdef _DEBUG_PARSE
		cout << "wrong section - no value returned!\n\n";
		#endif
		return(NULL);
	}	
	else {
		return(sectionItem);
	}
}

/* searches for a key in a specific section and sets its value to 'value'.
   if the section doesn't exist, it will be created. if the key doesn't
   exist yet, it will be created, too */
bool rcFile::setKey(const char *section, const char* key,
	const char *value)
{
	char sectionName[MAX_LINE_LENGTH+1];
	rcFileEntry *currentEntry=rootEntry;
	rcFileEntry *tmpEntryPointer=rootEntry;
	
	if (section[0]=='[') {
		sprintf(sectionName, "%s", section);
	}
	else {
		sprintf(sectionName, "[%s]", section);
	}
	
	/* search relevant section */
	while(currentEntry!=NULL) {
		if (strstr(currentEntry->key, sectionName)!=NULL) {
			if (currentEntry->type==tlSection) {
				#ifdef _DEBUG
				cout << "got section: \"" << section << "\"" << endl;
				#endif
				break;
			}
		}
		else {
			tmpEntryPointer=currentEntry; /* save last valid pointer */
			currentEntry=currentEntry->nextEntry;
		}
	}
	
	if (currentEntry==NULL) { /*section not found */
	
		/* create new section */
		if (rootEntry==NULL) {
			rootEntry=new rcFileEntry; /* create root entry */
			currentEntry=rootEntry;
		}
		else {
			tmpEntryPointer->nextEntry=new rcFileEntry; 
			currentEntry=tmpEntryPointer->nextEntry;
		}
		strcpy(currentEntry->key, sectionName); /* set section name */
		strcpy(currentEntry->value, "\0");
		currentEntry->type=tlSection;
		currentEntry->nextEntry=NULL;
		#ifdef _DEBUG
		cout << "created section: \"" << sectionName << "\"" << endl;
		#endif
		
		/* create section item */
		currentEntry->nextEntry=new rcFileEntry; /* alloc new key */
		currentEntry=currentEntry->nextEntry;
		strcpy(currentEntry->key, key); /* set key properties */
		strcpy(currentEntry->value, value);
		currentEntry->type=tlKey;
		currentEntry->nextEntry=NULL;
		#ifdef _DEBUG
		cout << "created key: \"" << key << "\", " << "value: \"" << value << "\"\n";
		#endif
		return(true);
	}
	
	/* section exists - now check if the key is already defined.
	   if defined change value */
	   
	tmpEntryPointer=currentEntry;   
	if (currentEntry->nextEntry!=NULL) /* get first line after section name */
		currentEntry=currentEntry->nextEntry;
   
	while(currentEntry!=NULL) {
		if (strcmp(currentEntry->key, key)==0) { /* if key exists */
			strcpy(currentEntry->value, value);
			#ifdef _DEBUG
			cout << "set key: \"" << key << "\", " << "value: \"" << value << "\"\n";
			#endif
			return(true);
		}
		else {
			if (currentEntry->type==tlSection) { /* new section entered */
				break;
			}
			tmpEntryPointer=currentEntry; /* save last valid pointer */
			currentEntry=currentEntry->nextEntry;
		}
	}

	/* create section item */
	currentEntry=tmpEntryPointer; /* get last valid pointer */
	tmpEntryPointer=currentEntry->nextEntry; /* save pointer for next */
	currentEntry->nextEntry=new rcFileEntry; /* alloc new key */
	currentEntry=currentEntry->nextEntry;
	strcpy(currentEntry->key, key); /* set key properties */
	strcpy(currentEntry->value, value);
	currentEntry->type=tlKey;
	currentEntry->nextEntry=tmpEntryPointer; /* put back pointer to next entry */
	#ifdef _DEBUG
	cout << "created key: \"" << key << "\", " << "value: \"" << value << "\"\n";
	#endif
	return(true);	
}

/* deletes a key out of a certain section. if afterwards section is empty, the
   section header will be removed,too.
   returns true if successful, otherwise false */
bool rcFile::delKey(const char* section, const char* key) 
{
	rcFileEntry *currentEntry, *tmpEntry;
	
	/* test if section exists */
	if ( (currentEntry=getSection(section))==NULL ) {
		return(false);
	}
	
	tmpEntry=currentEntry;
	currentEntry=currentEntry->nextEntry;
		
	/* now search item and delete it */
	while(1) {
		if (currentEntry==NULL) {
			#ifdef _DEBUG_PARSE
			cout << "delKey: key \"" << key << "\" not found!" << "\n\n";
			#endif
			return (false);
		}
		if (currentEntry->type==tlSection) {  /* entered new section */
			#ifdef _DEBUG_PARSE
			cout << "delKey: key \"" << key ;
			cout << "\" not found in section \"" << section << "\"!\n\n"; 
			#endif
			return(false);
		}
		if ((strcmp(currentEntry->key, key)==0)) {
			tmpEntry->nextEntry=currentEntry->nextEntry;
			if (currentEntry==cursor) {
				#ifdef _DEBUG
				cout << "delKey: reset cursor!" << endl;
				#endif
				cursor=rootEntry;
			}
			delete currentEntry;
			#ifdef _DEBUG
			cout << "delKey: deleted key \"" << key << "\"";
			cout << " in section \"" << section << "\"\n";
			#endif
			break;
		}
		else
			tmpEntry=currentEntry;
			currentEntry=currentEntry->nextEntry;
	}
	
	/* check if the last item in a section has been deleted.
	   eventually delete section header */
	if (countKeys(section)==0) {
		#ifdef _DEBUG
		cout << "delKey: section empty - deleting..." << endl;
		#endif
		purgeRcFile(false, false, false);	
	}
	return(true);	
}

/* modifies the name of a given section. returns true when successful, 
   otherwise false */
bool rcFile::setSection(const char *section, const char *newName) {

	rcFileEntry *currentEntry;

	if ((currentEntry=getSection(section))==NULL) {
		return(false);
	}
	else {
		if (section[0]=='[') {
			sprintf(currentEntry->key, "%s", newName);
		}
		else {
			sprintf(currentEntry->key, "[%s]", newName);
		}
	}
	return(true);
}

/* deletes given section. if onlyIfEmpty is true(default), it will only 
   be deleted if there are no keys inside the section. returns a boolean
   on wether deletion was successful or not. be aware that all empty
   sections in the list will be deleted on execution! also note that if the 
   cursor pointed to an entry in the deleted section, it will be reset to 
   point to the very first entry in the list */
bool rcFile::delSection(const char *section, bool onlyIfEmpty) {
	
	rcFileEntry *tmp;
	char *pKey;
	char key[MAX_LINE_LENGTH+1];
	
	tmp=cursor; /* save cursor */
	
	if (moveSection(section)==false) {
		cursor=tmp;
		return(false);
	}
	
	do {
		if ((pKey=moveNext())!=NULL) {
			strcpy(key, pKey);
		}
		else {
			break;
		}
		if (onlyIfEmpty) {
			#ifdef _DEBUG
			cout << "delSection: section \"" << section << "\" not empty!\n";
			#endif
			cursor=tmp; 
			return(false);
		}
		if (cursor==tmp) { /* if cursor is on deleted item */
			#ifdef _DEBUG
			cout << "delSection: reset cursor!" << endl;
			#endif
			tmp=rootEntry;
		}
		delKey(section, key);
	} while (moveSection(section));
	
	cursor=tmp;
	purgeRcFile(false, false, false);
	return(true);
}

/* returns the number of items found in a single section. When 0 is returned
   the section doesn't exist. */
int rcFile::countKeys(const char *section) {
	
	rcFileEntry *currentEntry;
	int qty=0;
	
	currentEntry=getSection(section);
	
	if (currentEntry==NULL) {	/* check if section exists */
		#ifdef _DEBUG
		cout << "countKeys: section not found:  \"" << section << "\"" << endl;
		#endif
		return(0);
	}
	currentEntry=currentEntry->nextEntry; /* step forward to first key */
	while (currentEntry!=NULL) {
		if (currentEntry->type==tlSection) {
			#ifdef _DEBUG
			cout << "countKeys: found " << qty << " keys in \"" << section << "\"" << endl;
			#endif
			return(qty);
		}
		else {
			if (currentEntry->type==tlKey) {	
				qty++;
			}
		}
		currentEntry=currentEntry->nextEntry;
	}
	return (qty);
}

/* returns the number of items found in a single section. When 0 is returned
   the section doesn't exist. */
int rcFile::countSections(void) {
	
	rcFileEntry *currentEntry;
	int qty=0;
	
	currentEntry=rootEntry;
	if (currentEntry==NULL) {	/* check if list is empty */
		#ifdef _DEBUG
		cout << "countSections: no sections found! " << endl;
		#endif
		return(0);
	}
	while (currentEntry!=NULL) {
		if (currentEntry->type==tlSection) {
			qty++;
		}
		currentEntry=currentEntry->nextEntry;
	}
	#ifdef _DEBUG
	cout << "countSections: found " << qty << " sections" << endl;
	#endif
	return (qty);
}

/* returns the number of all entries (sections and keys) in the list */
int rcFile::countEntries(void) {
	
	int qty=0;
	rcFileEntry *tmp=rootEntry;

	while (tmp!=NULL) {
		if ((tmp->type==tlKey) || (tmp->type==tlSection)) {
			qty++;
		}
		tmp=tmp->nextEntry;
	}
	#ifdef _DEBUG
	cout << "countKeys: found " << qty << " entries" << endl;
	#endif
	return(qty);
}

/* moves the cursor to the specified section.
   returns true if successful */
bool rcFile::moveSection(const char *section) {
	cursor=getSection(section);
	if (cursor==NULL) {
		#ifdef _DEBUG
		cout << "moveSection: no section \"" << section << "\"" << endl;
		#endif
		return(false);
	}
	else {
		#ifdef _DEBUG
		cout << "moveSection: \"" << section << "\"" << endl;
		#endif
		return (true);
	}
}

/* moves cursor to the next key found inside a section.
   returns the name of the key. if cursor is already pointing
   to the last key in that section, NULL will be returned and
   the cursor will remain at its position. */
char* rcFile::moveNext(void) {

	rcFileEntry *currentEntry=cursor;
	
	while (currentEntry!=NULL) {
		currentEntry=currentEntry->nextEntry;
		if (currentEntry==NULL) {
			#ifdef _DEBUG
			cout << "move Next: reached end of list!" << endl;
			#endif
			return(NULL);
		}
		if (currentEntry->type==tlSection) {
			#ifdef _DEBUG
			cout << "moveNext: cursor already on last key!" << endl;
			#endif
			return(NULL);
		}
		else {
			if (currentEntry->type==tlKey) {
				cursor=currentEntry;
				#ifdef _DEBUG
				cout << "moveNext: next key \"" << cursor->key << "\"" << endl;
				#endif
				return(cursor->key);
			}
			else {
				currentEntry=currentEntry->nextEntry;
			}
		}
	}
	return(NULL);
}

 /* returns false if either checksum is invalid or file is empty */
bool rcFile::isFileValid(void) {
	return(iIsValid);
}

/* opens file read-only or read-write-enabled depending on given parameter */
bool rcFile::openRcFile(bool readOnly) {
	if (readOnly) {
		iFile.open(iFileName.c_str(), ios::in);
	}
	else {
		iFile.open(iFileName.c_str(), ios::trunc | ios::out);
	}
	iIsOpen=iFile.good();
	return(iIsOpen);
}


/* closes file if it is open */
bool rcFile::closeRcFile(void) {
	iFile.close();
	iIsOpen=false;
	return(iFile.good());
}

/* read next line out of file */
char* rcFile::readNextItem(void) {
	
	static char item[MAX_LINE_LENGTH+1];

	if (iFile) {
		iFile.getline(item, MAX_LINE_LENGTH);
		if (!(iFile.rdstate() & ios::eofbit)) {
			return(item);
		}
		else {
			return(NULL);
		}
	}
	else
		return(NULL);
}

/* write next line to file */
void rcFile::writeNextItem(const char *entry) {
	iFile.write(entry, strlen(entry));
	iFile.write("\n", 1);
}

/* this is used to get a pointer to the rcFileEntry containing
   the given section-string (or NULL if not found) */
rcFileEntry* rcFile::getSection(const char *section) {
	
	rcFileEntry *currentEntry=rootEntry;
	char sectionName[MAX_LINE_LENGTH+1];
	
	if (section[0]=='[') {
		sprintf(sectionName, "%s", section);
	}
	else {
		sprintf(sectionName, "[%s]", section);
	}
	
	while(1) {
		if (currentEntry==NULL) { /* list is empty or reached end of list */
			#ifdef _DEBUG_PARSE
			cout << "get Section: section \"" << section << "\" not found!" << endl;
			#endif
			return (NULL);
		}
		if (currentEntry->type==tlSection) {
			if (strstr(currentEntry->key, sectionName)!=NULL) {
				#ifdef _DEBUG_PARSE
				cout << "getSection: got section: \"" << section << "\"" << endl;
				#endif
				break;
			}
			else
				currentEntry=currentEntry->nextEntry;
		}
		else
			currentEntry=currentEntry->nextEntry;
	}
	return(currentEntry);	
}

/* search a given key in the whole list. When a parameter for entry is given,
   the key is searched beginning with the item entry is pointing to.
   returns key's value or NULL if not found (or file couldn't be opened) */
char* rcFile::getItem(const char *key, const rcFileEntry *entry) {
	
	rcFileEntry *currentEntry=rootEntry;
	
	if (entry!=NULL) {
		currentEntry=(rcFileEntry*)entry;
	}
	iEnteredNewSection=false;

	while(1) {
		if (currentEntry==NULL) {
			#ifdef _DEBUG_PARSE
			cout << "getItem: key \"" << key << "\" not found!" << "\n\n";
			#endif
			return (NULL);
		}
		if (currentEntry->type==tlSection) {
			iEnteredNewSection=true;
		}	
		if (strcmp(currentEntry->key, key)==0) {
			#ifdef _DEBUG
			cout << "getItem: got key \"" << currentEntry->key << "\", ";
			cout << "value: \"" << currentEntry->value << "\"\n";
			#endif
			return (currentEntry->value);
		}
		else
			currentEntry=currentEntry->nextEntry;
	}
}

/* tries to extract key and value out of the given line 
   returns pointer to structure rcFileEntry containing the result.
   added: the line type will also be included in rcFileEntry */
rcFileEntry* rcFile::parseLine(const char *line) {
	
	char *value;
	int i=0, j=0;
	static rcFileEntry currentEntry;
	tLine currentType;
	 
	value=(char*)malloc((MAX_LINE_LENGTH+1)*sizeof(char));
	strcpy(value, line); /* line is const - therefore copy it */
	
	currentType=lineType(value);
	currentEntry.type=currentType;
	if (currentType!=tlKey) {
		strcpy(currentEntry.key, value);
		strcpy(currentEntry.value, "\0");
		
		#ifdef _DEBUG_PARSE
		cout << "\tparser: ";
		switch (currentType) {
			case tlSection:
				cout << "section: \"" << currentEntry.key << "\"";
				break;
			case tlComment:
				cout << "comment: \"" << currentEntry.key << "\"";
				break;
			case tlEmpty:
				cout << "empty line.";
				break;
			case tlUnknown:
				cout << "unknown: \"" << currentEntry.key << "\"";
				break;
			case tlKey:
				break;
		}
		cout << endl;		
		#endif
		return(&currentEntry);
	}
	
	/* now parse line to extract key and value */
	
	/* searches for a "=" */
	for (i=0; i<=(int)strlen(value); i++) {
		if (value[i]=='=')
			break;
	}
	
	/* get the last char of the key */
	for (j=i-1; j>0; j--) {
		if (isgraph(value[j])) 
			break;
		if (value[j] == '') 
			break;
		if (value[j] == '' || value[j]=='') 
			break;
		if (value[j] == '' || value[j]=='') 
			break;
		if (value[j] == '' || value[j]=='')
			break;
	}
	strncpy(currentEntry.key, value, j+1);
	currentEntry.key[j+1]='\0';
	i++;
	
	/* searches for the next printable char after "=" */
	for (j=i; j<(int)strlen(value); j++) {
		if (isgraph(value[j]))
			break;
		if (value[j] == '') 
			break;
		if (value[j] == '' || value[j]=='') 
			break;
		if (value[j] == '' || value[j]=='') 
			break;
		if (value[j] == '' || value[j]=='')
			break;
	}

	/* searches the end of the key value and terminates it */
	for (i=j; i<(int)strlen(value); i++) {
		if (!isgraph(value[i])) {
			if (value[j] == '') 
				continue;
			if (value[i] == '' || value[j]=='') 
				continue;
			if (value[i] == '' || value[j]=='') 
				continue;
			if (value[i] == '' || value[j]=='')
				continue;
			value[i]='\0';
			break;
		}
	}
	strcpy(currentEntry.value, &value[j]);
	#ifdef _DEBUG_PARSE
	cout << "\tparser: key: \"" << currentEntry.key <<"\", value: \""<< currentEntry.value << "\"\n";
	#endif
	return(&currentEntry);
}

/* tries to detect the type of a line - whether it contains a comment, a key 
   with value, a section header or is empty. Otherwise it's content will be
   declared to be unknown. */
tLine rcFile::lineType(const char *line) {
	
	int i;
	bool bMatch=false;
	
	/* check if it's a comment */
	for (i=0; i<=(int)strlen(line); i++) {
		if (isgraph(line[i])) {
			if ( line[i]=='#') {
				bMatch=true;
				break;
			}
			else {
				bMatch=false;
				break;
			}
		}
	}
	if (bMatch) {
		return(tlComment);
	}
		
	/* check if it's a section */
	bMatch=false;
	for (i=0; i<=(int)strlen(line); i++) {
		if (isgraph(line[i])) {
			if ( line[i]!='[') {
				bMatch=false;
				break;
			}
			if ( line[i]=='[') {
				for (i=(int)strlen(line); i>=0; i--) {
					if (isgraph(line[i])) {
						if ( line[i]!=']') {
							bMatch=false;
							break;
						}
						if ( line[i]==']') {		
							bMatch=true;
							break;
						}
					}
				}
			}
		}
	}
	if (bMatch) {
		return(tlSection);
	}
	
	
	/* searches for a "=". This implies the line to be a key line */
	bMatch=false;
	for (i=0; i<=(int)strlen(line); i++) {
		if (line[i]=='=') {
			bMatch=true;
			break;
		}
	}
	if (bMatch) {
		return(tlKey);
	}
	
	/* checks line to be empty */
	bMatch=false;
	for (i=0; i<=(int)strlen(line); i++) {
		if (isgraph(line[i])) {
			bMatch=false;
			break;
		}
		else {
			bMatch=true;
		}
	}
	if (bMatch) {
		return(tlEmpty);
	}
	
	/* otherwise the contents of the line couldn't be classified */
	return(tlUnknown);
}

/* calculates a checksum for the entry list */
char* rcFile::calcChecksum(void) {
	
	int i=0, cs=0, of=0;
	rcFileEntry *tmp=rootEntry;
	static char checksum[5];
	
	while (tmp!=NULL) {
		if (tmp->key[0]=='#') {	/* don't bother about comments */
			tmp=tmp->nextEntry;
			break;
		}
		for (i=0; i<(int)strlen(tmp->key); i++) {
			cs+=(tmp->key[i]<<8);
			of+=(cs >> 6);
			of=(of % 32768);
			cs=(cs % 32768);
		}
		for (i=0; i<(int)strlen(tmp->value); i++) {
			cs+=(tmp->value[i]<<8);
			of+=(cs >> 6);
			of=(of % 32768);
			cs=(cs % 32768);
		}
		cs+=of;
		cs=(cs % 32768);
		tmp=tmp->nextEntry;
	}
	checksum[4]='\0';
	for (i=3; i>=0; i--) {
		if ((of=cs%16)<=9) {
			checksum[i]=(char)(of+(int)'0');
		}
		else {
			of-=10;
			checksum[i]=(char)(of+(int)'A');
		}
		cs=cs>>4;
	}

	#ifdef _DEBUG	
	cout << "calculated cs: " << checksum << endl;
	#endif
	return(&checksum[0]);
}
